// Copyright (c) 2019 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.awt.event.*;
import java.io.*;

// This uses the CPM3.SYS file format, although code could be
// almost anything. Should be possible to put a CP/M 2.2 image
// in this format as comlen, and leave bnklen zero.
//
// NOTE: complications arise for OS versions that require
// files to complete booting, so truly diskless (CP/NOS) is
// not yet supported. Right now, a diskfull OS image could
// be booted over the network, provided the necessary files
// are still accessible once the OS is initialized. The merits
// of such a capability are debatable.
//
// For both CP/NET 1.2 and CP/NET 3, the OS must be fully
// initialized in order to load/start NDOS and SNIOS (CP/NET).
// Once we pass control to the CBOOT entry, we lose the
// ability to influence what gets executed. We can't ever
// call CBOOT once NDOS/SNIOS are running or those get
// obliterated.

// Basic semantics on Server:
//
// if msg[FMT] == NetworkBoot.FMT then
//     if nwb == null then
//         nwb = NetworkBoot.doStart(props, msg)
//         if nwb == null then
//             send(NetworkBoot.error(msg))
//             return
//         endif
//     endif
//     resetTimeout()
//     send(nwb.doNext(msg))
//     if nwb.done() then
//         nwb = null
//     endif
// endif
//
// timeout():
//     nwb = null
//

public class NetworkBoot {
	// route all messages with this FMT to us:
	public static final int FMT = 0xb0;

	static final int STATE_START = 0;
	static final int STATE_RESDMA = 1;
	static final int STATE_RESLD = 2;
	static final int STATE_BNKDMA = 3;
	static final int STATE_BNKLD = 4;
	static final int STATE_ORG0DMA = 5;
	static final int STATE_ORG0LD = 6;
	static final int STATE_DONE = 7;

	// CPM3.SYS file format:
	static final int MEMTOP = 0;
	static final int COMLEN = 1;
	static final int BNKTOP = 2;
	static final int BNKLEN = 3;
	static final int ENTRY = 4;	// 2 bytes
	static final int CNFTBL = 6;	// 2 bytes
	static final int COPY = 16;	// DRI Copyright message - use to detect ORG0
	static final int SIGNON = 128;	// '$' terminated string, 128 bytes max
	static final int IMAGE = 256;	// start of OS image (end header)

	static final int RECLEN = 128;

	private int state = -1;
	private byte[] image;
	private byte[] org0; // may be null
	private int curdma;
	private int reccnt;

	private int memtop;
	private int comlen;
	private int bnktop;
	private int bnklen;
	private int entry;
	private int cfgtbl;

	private byte[] resp;

	public NetworkBoot(byte[] img, byte[] img0, int sid) {
		image = img;
		org0 = img0;
		resp = new byte[128 + NetworkServer.DAT];
		memtop = (image[MEMTOP] & 0xff) << 8;	// 0000 mean 10000
		comlen = (image[COMLEN] & 0xff) << 8;
		bnktop = (image[BNKTOP] & 0xff) << 8;
		bnklen = (image[BNKLEN] & 0xff) << 8;
		entry = (image[ENTRY] & 0xff) | ((image[ENTRY+1] & 0xff) << 8);
		cfgtbl = (image[CNFTBL] & 0xff) | ((image[CNFTBL+1] & 0xff) << 8);
		state = STATE_START;
		if (cfgtbl != 0) {
			fixCfgTbl(sid);
		}
	}

	// If returns null, caller sends error(msg)
	public static NetworkBoot doStart(Properties props, byte[] msg) {
		// Can't get here unless FMT == B0
		if (msg[NetworkServer.FNC] != 1) {
			return null;
		}
		// since dir is used read-only, not so tight on security.
		String dir = props.getProperty("netboot_dir");
		if (dir == null) {
			dir = System.getProperty("user.home") + "/NetBoot";
		}
		String org0 = props.getProperty("netboot_org0");
		if (org0 == null) {
			org0 = "org0boot.bin";
		}
		String def = props.getProperty("netboot_default");
		if (def == null) {
			def = "defboot.sys";
		}
		int nid = (msg[NetworkServer.SID] & 0xff);
		int n = (msg[NetworkServer.SIZ] & 0xff) + 1;	// first byte LEN, last NUL
		String tag = "";
		File file = null;
		if (n > 2) {
			tag = new String(msg, NetworkServer.DAT + 1, n - 2).toLowerCase();
			// serach for a variation of filename that exists...
			file = new File(String.format("%s/cid%02X%s", dir, nid, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			if (!file.exists()) {
				file = new File(String.format("%s/cid%02X%s.sys", dir, nid, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
			if (!file.exists()) {
				file = new File(String.format("%s/%s", dir, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
			if (!file.exists()) {
				file = new File(String.format("%s/%s.sys", dir, tag));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
		} else {
			file = new File(String.format("%s/cid%02X.sys", dir, nid));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			if (!file.exists()) {
				file = new File(String.format("%s/%s", dir, def));
//System.err.format("Trying %s\n", file.getAbsolutePath());
			}
		}
		if (!file.exists()) {
System.err.format("No boot file \"%s\"\n", file.getAbsolutePath());
			return null;
		}
		byte[] img;
		byte[] img0 = null;
		try {
			FileInputStream f = new FileInputStream(file);
			img = new byte[f.available()];
			f.read(img);
			f.close();
		} catch (Exception ee) {
ee.printStackTrace();
			return null;
		}
		if (img[1] == 0 && img[3] == 0) {
System.err.format("No segments to load \"%s\"\n", file.getAbsolutePath());
			return null;
		}
		if ((img[COPY] & 0xff) == 'C') {
			File o0 = new File(dir + "/" + org0);
			if (!o0.exists()) {
				return null;
			}
			try {
				FileInputStream f = new FileInputStream(o0);
				img0 = new byte[f.available()];
				f.read(img0);
				f.close();
			} catch (Exception ee) {
ee.printStackTrace();
				return null;
			}
		}
		System.err.format("Client %02x booting file \"%s\" (ORG0 %s)\n",
			nid, file.getAbsolutePath(), (img0 != null));
		int sid = msg[NetworkServer.DID] & 0xff; // our ID
		return new NetworkBoot(img, img0, sid);
	}

	public static byte[] error(byte[] msg) {
		return new byte[]{ (byte)0xb1, msg[NetworkServer.SID],
					msg[NetworkServer.DID], 0, 0, 0 };
	}

	public boolean done() {
		return (state < 0 || state > STATE_DONE);
	}

	private void fixCfgTbl(int sid) {
		// cfgtbl must be within RES segment, and must not cross
		// record boundary.
		// copute distance from top - i.e. start of image
		int off = (memtop - (cfgtbl & 0xff80) - 128) & 0xffff;
		off |= (cfgtbl & 0x007f);
		off += IMAGE + 2;	// start of drive table
		for (int x = 0; x < 16; ++x) {
			int i = x * 2;
			if ((image[off + i] & 0x80) != 0 &&
				(image[off + i + 1] & 0xff) == 0xff) {
				image[off + i + 1] = (byte)(sid);
			}
		}
	}

	// Never returns null.
	public byte[] doNext(byte[] msg) {
		// Can't get here unless FMT == B0
		int adr;
		// TODO: validate SIZ, DAT?
		// TODO: FNC might be 1...
		if ((state != STATE_START && msg[NetworkServer.FNC] != 0) ||
			(state == STATE_START && msg[NetworkServer.FNC] != 1)) {
			state = -1;
		}
		switch (state) {
		case STATE_START:
			// TODO: does empty message require special?
			// TODO: scan for term, send only part used?
			resp[NetworkServer.FNC] = (byte)0x01;
			resp[NetworkServer.SIZ] = (byte)(128-1);
			System.arraycopy(image, 128, resp, NetworkServer.DAT, 128);
			if (comlen > 0) {
				state = STATE_RESDMA;
			} else {
				state = STATE_BNKDMA;
			}
			break;
		case STATE_RESDMA:	// start RES boot
			adr = (memtop - comlen) & 0xffff;
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			curdma = comlen + IMAGE - RECLEN;
			reccnt = comlen / RECLEN;
			state = STATE_RESLD;
			break;
		case STATE_RESLD: // send RES records
			// don't land here unless comlen > 0 (1+ records)
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			System.arraycopy(image, curdma, resp, NetworkServer.DAT, RECLEN);
			curdma -= RECLEN;
			--reccnt;
			if (reccnt <= 0) {
				if (bnklen > 0) {
					state = STATE_BNKDMA;
				} else if (org0 != null) {
					state = STATE_ORG0DMA;
				} else {
					state = STATE_DONE;
				}
			}
			break;
		case STATE_BNKDMA:	// start BNK boot
			adr = (bnktop - bnklen) & 0xffff;
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			curdma = (comlen + bnklen) + IMAGE - RECLEN;
			reccnt = bnklen / RECLEN;
			state = STATE_BNKLD;
			break;
		case STATE_BNKLD: // send BNK records
			// don't land here unless comlen > 0 (1+ records)
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			System.arraycopy(image, curdma, resp, NetworkServer.DAT, RECLEN);
			curdma -= RECLEN;
			--reccnt;
			if (reccnt <= 0) {
				if (org0 != null) {
					state = STATE_ORG0DMA;
				} else {
					state = STATE_DONE;
				}
			}
			break;
		case STATE_ORG0DMA:
			adr = 0x3000;
			resp[NetworkServer.FNC] = (byte)0x02;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)adr;
			resp[NetworkServer.DAT + 1] = (byte)(adr >> 8);
			org0[3] = (byte)entry;
			org0[4] = (byte)(entry >> 8);
			entry = adr;
			curdma = 0;
			// TODO: handle non-multiple of RECLEN...
			reccnt = org0.length / RECLEN;
			state = STATE_ORG0LD;
			break;
		case STATE_ORG0LD:
			resp[NetworkServer.FNC] = (byte)0x03;
			resp[NetworkServer.SIZ] = (byte)(RECLEN-1);
			System.arraycopy(org0, curdma, resp, NetworkServer.DAT, RECLEN);
			curdma += RECLEN;
			--reccnt;
			if (reccnt <= 0) {
				state = STATE_DONE;
			}
			break;
		case STATE_DONE: // done with boot.
			resp[NetworkServer.FNC] = (byte)0x04;
			resp[NetworkServer.SIZ] = (byte)(2-1);
			resp[NetworkServer.DAT] = (byte)entry;
			resp[NetworkServer.DAT + 1] = (byte)(entry >> 8);
			state = -1;
			break;
		default:	// error
			state = -1;
			return error(msg);
		}
		// if reach here, send a message...
		resp[NetworkServer.FMT] = (byte)0xb1;
		resp[NetworkServer.DID] = msg[NetworkServer.SID];
		resp[NetworkServer.SID] = msg[NetworkServer.DID];
		return resp;
	}
}
